@@ -20,6 +20,9 @@
 # SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 """\
+EasyCluster
+===========
+
 EasyCluster is a remote execution / clustering module for Python. 
 
 The latest version of this module is available from
@@ -68,6 +71,7 @@ from __future__ import print_function
 
 import sys
 import os
+import os.path
 import re
 import signal
 import socket
@@ -86,8 +90,8 @@ from types import GeneratorType, ModuleType, FunctionType, BuiltinFunctionType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.18.1'
-MIN_COMPAT_VERSION = '0.16'
+VERSION = '0.19'
+MIN_SERVER_VERSION = '0.09'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -637,8 +641,7 @@ class NonblockingResponse:
 
     def wait(self, origexc=None):
         '''Waits for the request to complete.
-
-    '''
+        '''
 
         if origexc is None:
             origexc = self._origexc
@@ -681,6 +684,8 @@ class Connection:
         self._enable_compression = enable_compression
         self._peer_enable_compression = False
         self._compress = self._decompress = None
+        self._peer_init_data = {}
+        self._delay_digest = b''
 
         self._init_data = {
             'version': VERSION,
@@ -736,6 +741,7 @@ class Connection:
         self._send_hmac = hmac.new(self._key, self._sendiv, HMAC_MOD)
         self._compress = zlib.compressobj(zlib.Z_BEST_COMPRESSION)
         self._decompress = zlib.decompressobj()
+        self._delay_digest = b''
         self._after_connect()
         
     def _after_connect(self):
@@ -747,7 +753,7 @@ class Connection:
     def _on_ready(self):
         pass
 
-    def send_message(self, typ, data):
+    def send_message(self, typ, data, delay_hmac=False):
         hmac = self._send_hmac
 
         with self._sendlock:
@@ -766,7 +772,19 @@ class Connection:
             else:
                 digest = hmac.digest()
                 hmac.update(digest)
-            _sock_sendall(self._sock, b''.join((hdr, data, digest)))
+
+            prev_hmac = self._delay_digest
+            if prev_hmac:
+                if typ == ERROR:
+                    prev_hmac = b'\x00' * len(prev_hmac)
+                self._delay_digest = b''
+
+            if delay_hmac:
+                data = b''.join((prev_hmac, hdr, data))
+                self._delay_digest = digest
+            else:
+                data = b''.join((prev_hmac, hdr, data, digest))
+            _sock_sendall(self._sock, data)
 
     def make_pickle(self, object):
         sio = BytesIO()
@@ -786,26 +804,19 @@ class Connection:
         return True
 
     def _got_init(self, p):
-        self._peer_version = p.get('version')
+        self._peer_version = vers = p.get('version')
+        self._peer_init_data = p
+        
         self._peer_enable_compression = p.get('enable-compression', False)
         
-        min_vers_tuple = tuple(int(v) for v in MIN_COMPAT_VERSION.split('.'))
-        peer_vers_tuple = tuple(int(v) for v in self._peer_version.split('.'))
-        if peer_vers_tuple < min_vers_tuple:
-            svr_vers = VERSION
-            client_vers = self._peer_version
-            me = 'server'
-            if isinstance(self, Client):
-                svr_vers, client_vers = client_vers, svr_vers
-                me = 'client'
-
-            raise ValueError('Incompatible EasyCluster version: client has version %s, server has version %s; '
-                             '%s requires at least version %s' % (client_vers, svr_vers, me, MIN_COMPAT_VERSION))
-        return True
+        return self._check_version(vers)
 
     def _got_keepalive(self):
         pass
 
+    def _got_challenge(self, data):
+        self.send_message(CHALLENGE_RESPONSE, data)
+
     def _read_data(self, block=True):
         ## attribute access protected by GIL
         sock = self._sock
@@ -832,6 +843,9 @@ class Connection:
             recv_buf.write(data)
             newlen = self._recv_expect - len(data)
             if newlen > 0:
+                if newlen > 256 and not self._verified:
+                    self.send_message(ERROR, b'Initial message too long')
+                    self.close()
                 self._recv_expect = newlen
                 continue
 
@@ -886,17 +900,18 @@ class Connection:
                         return self._got_response(pick)
                     elif rmsg == INIT:
                         try:
-                            rv = self._got_init(pick)
+                            ready = self._got_init(pick)
                         except ValueError as e:
                             self.send_message(ERROR, _str_to_bytes(e))
                             raise
-                        self._on_ready()
-                        return rv
+                        if ready:
+                            self._on_ready()
+                        return True
 
                 elif rmsg == KEEPALIVE:
                     self._got_keepalive()
                 elif rmsg == CHALLENGE:
-                    self.send_message(CHALLENGE_RESPONSE, rdata)
+                    self._got_challenge(rdata)
                 elif rmsg == CHALLENGE_RESPONSE:
                     if rdata == self._my_nonce:
                         self._verified = True
@@ -940,6 +955,45 @@ class _GlobalProxy:
     def __call__(self, *args, **kw):
         return self._client.raw_call(self._name, args, kw, *_get_special_keywords(kw))
 
+_upgrade_code = r'''
+import sys, os, zlib
+import easycluster as _core
+import easycluster.server as _server
+def upgrade(osvr, defmod):
+    global coredata, svrdata, upgradecode
+    for mod, ztxt in ((_core, coredata), (_server, svrdata)):
+        cb = compile(zlib.decompress(ztxt).decode('utf-8'), mod.__file__, 'exec', dont_inherit=True)
+        eval(cb, mod.__dict__)
+    defmod.define('import easycluster\nfrom easycluster import *\n')
+    return _server.Server(osvr._key, osvr._sock, osvr._peername, osvr._verbosity, osvr)
+''' # '''
+
+## Bootstrap code to do an in-memory upgrade on versions older than 0.19
+_compat_upgrade_code = _upgrade_code + r'''
+defmod = sys.modules[initdata['definitions_module']]
+osvr = defmod.running_server
+def hack_req(*args):
+    osvr._peer_init_data = initdata
+    if type(osvr).__name__ == 'QuietServer':
+        osvr._verbosity = 0
+    else:
+        osvr._verbosity = 2
+    nsvr = upgrade(osvr, defmod)
+    nsvr._got_request(*args)
+    _exit = os._exit
+    globals().clear()
+    try:
+        while nsvr:
+            nsvr = nsvr.run()
+    except Exception:
+        import traceback
+        traceback.print_exc()
+    finally:
+        _exit(0)
+
+osvr._got_request = hack_req
+''' # '''
+
 
 class Client(Connection):
     '''Represents the client side of the connection.'''
@@ -1012,6 +1066,48 @@ class Client(Connection):
         params = parse_connection_spec(spec, default_port, default_key, (), warn)
         return cls(**params)
 
+    def _send_upgrade(self, compat=False):
+        '''Sends upgrade code to the remote host'''
+        coref = os.path.splitext(__file__)[0] + '.py'
+        svrf = os.path.join(os.path.dirname(coref), 'server.py')
+        coretxt, svrtxt = [zlib.compress(open(f, 'rb').read(), 9) for f in (coref, svrf)]
+        data = {
+            'coredata': coretxt,
+            'svrdata': svrtxt,
+        }
+        kw = {}
+        if compat:
+            data['initdata'] = self._init_data
+            if PYTHON3:
+                call = ('exec', (_compat_upgrade_code, data), kw)
+            else:
+                data['upgradecode'] = _compat_upgrade_code
+                eval_args = ("eval(compile(upgradecode,'upgrade','exec'))", data)
+                call = ('eval', eval_args, kw)
+            request = (-1, [], SINGLE, 0, 'call', call, kw)
+        else:
+            request = (-1, [], None, 0, 'do_easycluster_upgrade', (_upgrade_code, data), kw)
+        p = self.make_pickle(request)
+        self._completion[-1] = self._upgrade_complete, self._init_error
+        self.send_message(REQUEST, p)
+        return False
+
+    def _upgrade_complete(self, resp):
+        self._on_ready()
+
+    def _check_version(self, pv):
+        if _version_compare(pv, VERSION) != 0:
+            print('me = %s o = %s' % (VERSION, pv))
+            if _version_compare(pv, '0.09') < 0:
+                raise ValueError('Incompatible EasyCluster version: client has version %s, server has version %s; '
+                                 'client requires server to have at least version %s' % (VERSION, pv, MIN_SERVER_VERSION))
+
+            else:
+                return self._send_upgrade(_version_compare(pv, '0.19') < 0)
+            ## Now we know the server supports compression.
+            self._peer_enable_compression = False
+        return True
+
     def _init_thread(self):
         self.send_challenge()
 
@@ -1272,8 +1368,8 @@ class Client(Connection):
     def call(self, _func, *args, **kwargs):
         '''Call a function by name on the remote server.
 
-        @type  func: str
-        @param func: The absolute name of the function to call.
+        @type  _func: str
+        @param _func: The absolute name of the function to call.
 
         See L{Client.send_request} for the meaning of the C{nonblocking},
         C{oncomplete}, C{onerror}, C{threadid}, and C{origexc} arguments.
@@ -1338,7 +1434,7 @@ class Client(Connection):
 
     @property
     def server_version(self):
-        '''Return the version of EasyCluster on the server'''
+        '''Return the actual version of EasyCluster on the server.'''
         return self._peer_version
 
     def __getattr__(self, attr):
@@ -1491,6 +1587,7 @@ class _ATCThread(threading.Thread):
                         ## Check thread status after every valid message received
                         atc.check_threads()
         except (EnvironmentError, select.error) as e:
+            closeargs = (e,)
             if e.args[0] != errno.ECONNRESET and e.args[0] != errno.EBADF:
                 raise
         except EOFError:
@@ -1503,7 +1600,7 @@ class _ATCThread(threading.Thread):
         finally:
             atc.close(*closeargs)
 
-class ThreadedClient(Client, threading.Thread):
+class ThreadedClient(Client):
     '''A client that automatically creates unique threads on the server for each
     thread on the client that makes remote calls.
 
@@ -2127,6 +2224,18 @@ def _add_extra_classes():
 _add_extra_classes()
 del _add_extra_classes
 
+def _vers_tuple(vers):
+    return tuple(int(v) for v in vers.split('.'))
+
+def _version_compare(va, vb):
+    va = _vers_tuple(va)
+    vb = _vers_tuple(vb)
+    if va < vb:
+        return -1
+    if va > vb:
+        return 1
+    return 0
+
 if sys.platform.startswith('win'):
     from ctypes import windll, WinError, wintypes
     kernel32 = windll.kernel32
