@@ -24,6 +24,8 @@ from __future__ import print_function
 
 import sys
 
+MIN_CLIENT_VERSION = '0.16'
+
 # A little bit of magic to allow running python -m easycluster.server. This
 # allows the user to use '-m' with Python 2.6, which fails when '-m' is used on
 # a package.
@@ -194,7 +196,13 @@ class RootObject(object):
     def import_modules(self, names):
         for n in names:
             setattr(self._defmod, n, __import__(n))
-        
+
+    def do_easycluster_upgrade(self, code, data):
+        cb = compile(code, 'upgrade', 'exec')
+        eval(cb, data)
+        new = data['upgrade'](self._defmod.running_server, self._defmod)
+        raise _Replace(new)
+
 def profile_request(func, args):
     '''Profile a function call and write the results to the user's
     home directory. Used only for debugging.'''
@@ -234,13 +242,17 @@ class ServerThread(threading.Thread):
             #profile_request(svr.run_request, (rid, obj, meth, args, kw))
             svr.run_request(rid, obj, meth, args, kw)
 
+class _Replace(BaseException):
+    def __init__(self, new):
+        self.new = new
+
 class Server(_core.Connection):
     '''Represents the server side of a remote connection.'''
 
     _sendiv = b'SRVR'
     _recviv = b'CLNT'
 
-    def __init__(self, key, sock, peername):
+    def __init__(self, key, sock, peername, verbosity=0, repl=None):
         _core.Connection.__init__(self, key, True)
         self._next_object_id = 1
         self._next_type_id = 1
@@ -250,32 +262,55 @@ class Server(_core.Connection):
         self._root = root
         self._local_object_by_id = {0: root}
         self._local_id_by_object = {id(root): 0}
-        #self._local_typedat_by_id = {}
         self._local_id_by_typedat = {}
         
         self._pending_responses = deque()
         self._keep_alive_interval = None
-
-        self._threads = {}
-        self.set_socket(sock, peername)
-
         
+        self._threads = {}
+        self._verbosity = verbosity
+        self._sent_challenge = False
+
+        if repl:
+            self._sent_challenge = True
+
+        self.set_socket(sock, peername)
+        if repl:
+            self._verified = True
+            self._recv_hmac = repl._recv_hmac
+            self._send_hmac = repl._send_hmac
+            self._remote_host = repl._remote_host
+            self._remote_port = repl._remote_port
+            self._got_init(repl._peer_init_data)
+            
+    def report_connection(self):
+        '''Report that a connection was received.'''
+        if self._verbosity >= 1:
+            print('New connection from %s' % (self._peername))
+
+    def report_disconnection(self):
+        '''Report that a client disconnected.'''
+        if self._verbosity >= 1:
+            print('Connection to %s terminated' % (self._peername))
+
+    def report_request(self, rid, garbage, tid, oid, meth, args, kw):
+        '''Report a request from the client.'''
+        if self._verbosity >= 2 and rid is not None:
+            print('<%s> Request %d: <%s> %d.%s g=%s' % (self._peername, rid, tid, oid, meth, garbage))
+
     def _after_connect(self):
         self._sock.setblocking(1)
         self.send_challenge()
 
-    def report_connection(self):
-        '''Report that a connection was received.'''
-        print('New connection from %s' % (self._peername))
+    def _check_version(self, pv):
+        if _core._version_compare(pv, MIN_CLIENT_VERSION) < 0:
+            raise ValueError('Incompatible EasyCluster version: client has version %s, server has version %s; '
+                             'server requires client to have at least version %s' % (pv, VERSION, MIN_CLIENT_VERSION))
 
-    def report_disconnection(self):
-        '''Report that a client disconnected.'''
-        print('Connection to %s terminated' % (self._peername))
-
-    def report_request(self, rid, garbage, tid, oid, meth, args, kw):
-        '''Report a request from the client.'''
-        if rid is not None:
-            print('Request %d: <%s> %d.%s g=%s' % (rid, tid, oid, meth, garbage))
+    def send_challenge(self):
+        if not self._sent_challenge:
+            self.send_message(_core.CHALLENGE, self._my_nonce, True)
+            self._sent_challenge = True
 
     def run_request(self, rid, oid, meth, args, kw):
         '''Processes a single request.'''
@@ -284,6 +319,10 @@ class Server(_core.Connection):
             rval = getattr(obj, meth)(*args, **kw)
             with self._lock:
                 rpickle = self.make_pickle((rid, rval))
+        except _Replace:
+            rpickle = self.make_pickle((rid, None))
+            self.send_message(_core.RESPONSE, rpickle)
+            raise
         except Exception:
             typ, val, tb = sys.exc_info()
             text = traceback.format_exception(typ, val, tb)
@@ -314,8 +353,11 @@ class Server(_core.Connection):
             if t is not None:
                 del self._threads[tid]
         elif t is None:
-            self._threads[tid] = t = ServerThread(self, tid)
-            t.start()
+            if tid is None:
+                self.run_request(rid, oid, meth, args, kw)
+            else:
+                self._threads[tid] = t = ServerThread(self, tid)
+                t.start()
 
         if t is not None:
             t.queue.put((rid, oid, meth, args, kw))
@@ -355,6 +397,9 @@ class Server(_core.Connection):
                 raise
         except EOFError:
             pass
+        except _Replace as r:
+            return r.new
+
         self.report_disconnection()
 
         threads = self._threads.values()
@@ -391,20 +436,10 @@ class Server(_core.Connection):
         tid = self._next_type_id
         self._next_type_id = tid + 1
         self._local_id_by_typedat[typedat] = tid
-        #self._local_typedat_by_id[tid] = typ
-        #print('typedat for %s (%d) is %r' % (typ, tid, typedat))
         return '%d\n%d\n%s' % (oid, tid, typedat)
 
-class QuietServer(Server):
-    '''A subclass of Server which does not print anything to output.'''
-    def report_connection(self):
-        pass
-
-    def report_disconnection(self):
-        pass
-
-    def report_request(self, *args):
-        pass
+## For compatibility
+QuietServer = Server
 
 class IPAddress(object):
     '''Represents an IP address with optional network mask length'''
@@ -479,13 +514,13 @@ class IPAddress(object):
         return (self.family == net.family and 
                 (self.aslong & net.mask) == net.aslong)
 
-def spawn_local(clientclass=_core.Client, svrclass=QuietServer):
+def spawn_local(clientclass=_core.Client, server_verbosity=0):
     '''Create a private server instance and connect to it.'''
     global _local_id
     csock, ssock = socketpair()
     _core._mark_non_inheritable(csock.fileno())
 
-    spawn_runner_process(ssock, 'local:%d' % _local_id, svrclass, b'')
+    spawn_runner_process(ssock, 'local:%d' % _local_id, b'', server_verbosity)
     ssock.close()
 
     rc = clientclass(b'')
@@ -506,23 +541,26 @@ def server_main(args=None):
     else:
         servmod = None
 
-    from optparse import OptionParser
+    from optparse import OptionParser, SUPPRESS_HELP
     if args is None:
         args = sys.argv[1:]
 
     key = ''
     options = OptionParser(description="Runs the EasyCluster service and generates keyfiles", version=_core.VERSION)
     _core.add_key_options(options)
+    options.add_option('-v', '--verbose', default=0, action='count', help='Increase the verbosity of the server.')
     options.add_option('-S', '--serve', metavar='FILE', action='store_true', help='Run the server on the specified port')
     options.add_option('-g', '--generate', metavar='FILE', help='Generate random HMAC key and '
                        'write it to FILE. The file will be created with read access only for the owner.')
     options.add_option('-O', '--overwrite', action='store_true', help='Allow --generate to overwrite existing keyfile')
     options.add_option('-p', '--port', type='int', default=_core.DEFAULT_PORT, help='TCP port to listen on. Default: %d' % _core.DEFAULT_PORT)
-    options.add_option('-c', '--class', dest='svrclass', default='Server', help='Fully-qualified name of server class')
     options.add_option('-b', '--bind', default=None, help='Bind the server to a specific IP address')
     options.add_option('-a', '--allow', action='append', default=[], help='Allow only a specific address or subnet to connect '
                        '(e.g. 192.168.0.0/24). Specify more than once to allow multiple addresses.')
 
+    ## Deprecated. Generally used only to specify QuietServer - quiet operation is now the default.
+    options.add_option('-c', '--class', dest='svrclass', default=None, help=SUPPRESS_HELP)
+
     if _daemonize:
         options.add_option('-d', '--daemonize', action='store_true',
                            help='Detach from current terminal and run in the background')
@@ -592,26 +630,41 @@ def server_main(args=None):
         key = _core.key_from_options(opts)
         if key == b'':
             print('SECURITY WARNING: running with blank key.', file=sys.stderr)
-        if '.' in opts.svrclass:
-            svrmod, svrclassname = opts.svrclass.rsplit('.')
-            clas = getattr(_core._get_module(svrmod), svrclassname)
-        else:
-            clas = globals()[opts.svrclass]
 
-        run_server(opts.port, key, opts.bind, opts.allow, clas,
+        run_server(opts.port, key, opts.bind, opts.allow, opts.verbose,
                    getattr(opts, 'daemonize', False),
                    getattr(opts, 'pidfile', None))
     else:
         options.print_help()
 
 def run_server(port, key, bindaddr=None, allow=(), 
-               svrclass=Server, daemonize=False, pidfile=None):
+               verbosity=0, daemonize=False, pidfile=None):
     '''Runs a server on the specified port.
 
-    port is the TCP port to listen on.
-    key is the HMAC key to use.
-    svrclass is the server class to instantiate for each client
-    (default: easycluster.Server).
+    @type  port: int
+    @param port: The TCP port to listen on.
+
+    @type  key: bytes
+    @param key: Binary HMAC key to use.
+
+    @type  bindaddr: str
+    @param bindaddr: Interface address to bind on, or None to bind on all
+    interfaces.
+
+    @type  allow: sequence
+    @param allow: Sequence of IP networks to allow connections from, or empty
+    list to allow connections from all IP addresses.
+    
+    @type  verbosity: int
+    @param verbosity: Level of reporting: 0 = no reporting, 1 = report
+    connections, 2 = report connections and requests.
+    
+    @type  daemonize: bool
+    @param daemonize: If True, on POSIX systems, become a daemon after starting
+    server. Not supported on Windows
+
+    @type  pidfile: str
+    @param pidfile: Path to file to write process ID to when daemonizing.
     '''
     global _server_socket
 
@@ -678,7 +731,7 @@ def run_server(port, key, bindaddr=None, allow=(),
                     print('Connection from %s not allowed' % addr[0])
                     continue
 
-            spawn_runner_process(csock, _core._get_peer_name(addr), svrclass, key)
+            spawn_runner_process(csock, _core._get_peer_name(addr), key, verbosity)
         except Exception:
             traceback.print_exc()
         finally:
@@ -778,23 +831,25 @@ if sys.platform.startswith('win'):
     
     MAGIC_ARG = '--easycluster-run-server'
 
-    def spawn_runner_process(csock, peername, svrclass, key):
+    def spawn_runner_process(csock, peername, key, verbosity):
         args = [_python_exe, '-m', 'easycluster.__main__', MAGIC_ARG]
         p = subprocess.Popen(args, stdin=subprocess.PIPE)
         pinfo = dupsock(csock, p.pid)
         p.stdin.write(_core.b2a_hex(
                 _core.pickle.dumps((pinfo, csock.family, csock.type,
-                                    csock.proto, peername, svrclass, key),
+                                    csock.proto, peername, key, verbosity),
                                    protocol=_core.pickle.HIGHEST_PROTOCOL)) + b'\r\n')
         p.stdin.close()
 
     def server_check_runner():
         if len(sys.argv) == 2 and sys.argv[1] == MAGIC_ARG:
-            pinfo, family, type, proto, peername, svrclass, key = _core.pickle.loads(_core._hexdec(sys.stdin.readline().strip()))
+            pinfo, family, type, proto, peername, key, verbosity = _core.pickle.loads(_core._hexdec(sys.stdin.readline().strip()))
             sys.stdin.close()
             sock = make_socket(family, type, proto, pinfo)
             try:
-                svrclass(key, sock, peername).run()
+                svr = Server(key, sock, peername, verbosity)
+                while svr:
+                    svr = svr.run()
             except Exception:
                 traceback.print_exc()
             sys.exit(0)
@@ -815,14 +870,16 @@ else:
     def server_check_runner():
         pass
 
-    def spawn_runner_process(csock, peername, svrclass, key):
+    def spawn_runner_process(csock, peername, key, verbosity):
         cpid = os.fork()
         if cpid == 0:
             fd = csock.fileno()
             os.closerange(3, fd)
             os.closerange(fd + 1, _max_fd)
             try:
-                svrclass(key, csock, peername).run()
+                svr = Server(key, csock, peername, verbosity)
+                while svr:
+                    svr = svr.run()
             except Exception:
                 traceback.print_exc()
             sys.exit(0)
