@@ -365,7 +365,7 @@ def parallel3():
 @define_test('Test spawnlocal')
 def testlocal():
     global l1
-    l1 = spawn_local(svrclass=easycluster.server.QuietServer)
+    l1 = spawn_local()
     reftest()
     methods()
     attrs()
